/*___________________________________________________________________________
 |                                                                           |
 | Fixed-point C simulation of AMR WB ACELP coding algorithm with 20 ms      |
 | speech frames for wideband speech signals.                                |
 |___________________________________________________________________________|
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "typedef.h"
#include "basic_op.h"
#include "acelp.h"
#include "cnst.h"
#include "main.h"
#include "bits.h"
#include "count.h"
#include "cod_main.h"


/*-----------------------------------------------------------------*
 * CODER.C                                                         *
 * ~~~~~~~~~~                                                      *
 * Main program of the AMR WB ACELP wideband coder.                *
 *                                                                 *
 *    Usage : coder (-dtx) mode speech_file  bitstream_file		   *
 *                                                                 *
 *    Format for speech_file:                                      *
 *      Speech is read from a binary file of 16 bits data.         *
 *                                                                 *
 *    Format for bitstream_file:                                   *
 *        1 word (2-byte) for the type of frame type			   *
 *          (TX_FRAME_TYPE or RX_FRAME_TYPE) 					   *
 *        1 word (2-byte) for the frame type					   *
 *          (see dtx.h for possible values)						   *
 *        1 word (2-byte) for the mode indication				   *
 *          (see bits.h for possible values)					   *
 *		  N words (2-byte) containning N bits.                     *
 *          Bit 0 = 0x0000 and Bit 1 = 0x0001                      *
 *                                                                 *
 *    mode = 0..8 (bit rate = 6.60 to 23.85 k)                     *
 *                                                                 *
 *    -dtx if DTX is ON			                                   *
 *-----------------------------------------------------------------*/

int main(int argc, char *argv[])
{
    FILE *f_speech;                        /* File of speech data                   */
    FILE *f_serial;                        /* File of serial bits for transmission  */
    FILE *f_mode = NULL;                   /* File of modes for each frame			 */

    Word16 signal[L_FRAME16k];             /* Buffer for speech @ 16kHz             */
    Word16 prms[NB_BITS_MAX];

    Word16 coding_mode, nb_bits, allow_dtx, mode_file, mode = 0, i;
    Word16 reset_flag;
    long frame;
    char Mode[2] = "0";

    void *st;
    TX_State *tx_state;

#ifdef ITU_interface
    Word16 quiet;

    quiet = 0;
    if(argc == 5 || argc == 6)
    {
        if(strcmp(argv[1], "-q") == 0)
        {
            quiet = 1;
            argv++;
        }
    }
    fprintf(stderr, "\n");
    fprintf(stderr, " ========================================================================\n");
    fprintf(stderr, " 3GPP/ITU-T AMR Wideband Codec, November 30, 2001. Version %s.\n", CODEC_VERSION);
    fprintf(stderr, " (ITU-T interface).\n");
    fprintf(stderr, " ========================================================================\n");
    fprintf(stderr, "\n");
    if ((argc < 4) || (argc > 6))
    {
        fprintf(stderr, "Usage : encoder  (-q) mode speech_file  bitstream_file\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for speech_file:\n");
        fprintf(stderr, "  Speech is read form a binary file of 16 bits data.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for bitstream_file:\n");

        fprintf(stderr, "  One sync word (2-byte) (0x6b21)\n");
        fprintf(stderr, "  One word (2-byte) to indicate data length.\n");
        fprintf(stderr, "  N words (2-byte) containing N data bits.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "mode: 0 to 8 (9 bits rates)\n");
       fprintf(stderr, " ===================================================================\n");
        fprintf(stderr, " mode   :  (0)  (1)   (2)   (3)   (4)   (5)   (6)   (7)   (8)     \n");
        fprintf(stderr, " bitrate: 6.60 8.85 12.65 14.25 15.85 18.25 19.85 23.05 23.85 kbit/s\n");
        fprintf(stderr, " ===================================================================\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "-q is quiet operation\n");
        fprintf(stderr, "\n");
        exit(0);
    }

    allow_dtx = 0;

    mode_file = 0;
    if (strcmp(argv[1], "-modefile") == 0)
    {
        mode_file = 1;
        argv++;
        if ((f_mode = fopen(argv[1], "r")) == NULL)
        {
            fprintf(stderr, "Error opening input file  %s !!\n", argv[1]);
            exit(0);
        }
        fprintf(stderr, "Mode file:  %s\n", argv[1]);
    } else
	{
		mode = (Word16) atoi(argv[1]);
    	if ((mode < 0) || (mode > 8))
    	{
    	   fprintf(stderr, " error in bit rate mode %d: use 0 to 8\n", mode);
    	   exit(0);
    	} else
    	   nb_bits = nb_of_bits[mode];
   	}

#else

    fprintf(stderr, "\n");
    fprintf(stderr, " ==========================================================================\n");
    fprintf(stderr, " 3GPP/ITU-T AMR Wideband Codec, November 30, 2001. Version %s.\n", CODEC_VERSION);
    fprintf(stderr, " ==========================================================================\n");
    fprintf(stderr, "\n");

    /*-------------------------------------------------------------------------*
     * Open speech file and result file (output serial bit stream)             *
     *-------------------------------------------------------------------------*/

    if ((argc < 4) || (argc > 6))
    {
        fprintf(stderr, "Usage : coder  (-dtx) mode speech_file  bitstream_file\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for speech_file:\n");
        fprintf(stderr, "  Speech is read form a binary file of 16 bits data.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for bitstream_file:\n");
        fprintf(stderr, "  One word (2-byte) to indicate type of frame type.\n");
        fprintf(stderr, "  One word (2-byte) to indicate frame type.\n");
        fprintf(stderr, "  One word (2-byte) to indicate mode.\n");
        fprintf(stderr, "  N words (2-byte) containing N bits.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "mode: 0 to 8 (9 bits rates) or\n");
        fprintf(stderr, "      -modefile filename\n");
        fprintf(stderr, " ===================================================================\n");
        fprintf(stderr, " mode   :  (0)  (1)   (2)   (3)   (4)   (5)   (6)   (7)   (8)     \n");
        fprintf(stderr, " bitrate: 6.60 8.85 12.65 14.25 15.85 18.25 19.85 23.05 23.85 kbit/s\n");
        fprintf(stderr, " ===================================================================\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "-dtx if DTX is ON, default is OFF\n");
        fprintf(stderr, "\n");
        exit(0);
    }
    allow_dtx = 0;
    if (strcmp(argv[1], "-dtx") == 0)
    {
        allow_dtx = 1;
        argv++;
    }
    mode_file = 0;
    if (strcmp(argv[1], "-modefile") == 0)
    {
        mode_file = 1;
        argv++;
        if ((f_mode = fopen(argv[1], "r")) == NULL)
        {
            fprintf(stderr, "Error opening input file  %s !!\n", argv[1]);
            exit(0);
        }
        fprintf(stderr, "Mode file:  %s\n", argv[1]);
    } else
    {
        mode = (Word16) atoi(argv[1]);
        if ((mode < 0) || (mode > 8))
        {
            fprintf(stderr, " error in bit rate mode %d: use 0 to 8\n", mode);
            exit(0);
        } else
            nb_bits = nb_of_bits[mode];
    }

#endif

    if ((f_speech = fopen(argv[2], "rb")) == NULL)
    {
        fprintf(stderr, "Error opening input file  %s !!\n", argv[2]);
        exit(0);
    }
#ifdef ITU_interface
        if(quiet == 0)
#endif
    fprintf(stderr, "Input speech file:  %s\n", argv[2]);

    if ((f_serial = fopen(argv[3], "wb")) == NULL)
    {
        fprintf(stderr, "Error opening output bitstream file %s !!\n", argv[3]);
        exit(0);
    }
#ifdef ITU_interface
        if(quiet == 0)
#endif
    fprintf(stderr, "Output bitstream file:  %s\n", argv[3]);

    /*-------------------------------------------------------------------------*
     * Initialisation                                                          *
     *-------------------------------------------------------------------------*/

    Init_coder(&st);                       /* Initialize the coder */
    Init_write_serial(&tx_state);
    Init_WMOPS_counter();                  /* for complexity calculation */

    /*---------------------------------------------------------------------------*
     * Loop for every analysis/transmission frame.                               *
     *   -New L_FRAME data are read. (L_FRAME = number of speech data per frame) *
     *   -Conversion of the speech data from 16 bit integer to real              *
     *   -Call coder to encode the speech.                                       *
     *   -The compressed serial output stream is written to a file.              *
     *--------------------------------------------------------------------------*/


#ifdef ITU_interface
        if(quiet == 0)
#endif
    fprintf(stderr, "\n --- Running ---\n");

    frame = 0;

    while (fread(signal, sizeof(Word16), L_FRAME16k, f_speech) == L_FRAME16k)
    {
        Reset_WMOPS_counter();

        if (mode_file)
        {
            if (fread(Mode, sizeof(char), 1, f_mode) != 1)
            {
                fprintf(stderr, "\nend of mode control file reached\n");
                fprintf(stderr, "From now on using mode: %d kbit/s.\n", nb_of_bits[mode]);
            }
            mode = (Word16) atoi(Mode);
        }
        coding_mode = mode;

        frame++;

#ifdef ITU_interface
        if(quiet == 0)
#endif
        fprintf(stderr, " Frames processed: %ld\r", frame);

        /* check for homing frame */
        reset_flag = encoder_homing_frame_test(signal);

        for (i = 0; i < L_FRAME16k; i++)   /* Delete the 2 LSBs (14-bit input) */
        {
            signal[i] = (Word16) (signal[i] & 0xfffC);      logic16(); move16();
        }

        coder(&coding_mode, signal, prms, &nb_bits, st, allow_dtx);

        Write_serial(f_serial, prms, coding_mode, mode, tx_state);

        WMOPS_output((Word16) (coding_mode == MRDTX));

        /* perform homing if homing frame was detected at encoder input */
        if (reset_flag != 0)
        {
            Reset_encoder(st, 1);
        }
    }

    Close_coder(st);

    exit(0);
}
