/*___________________________________________________________________________
 |                                                                           |
 | Fixed-point C simulation of AMR WB ACELP coding algorithm with 20 ms      |
 | speech frames for wideband speech signals.                                |
 |___________________________________________________________________________|
*/

#include <stdio.h>
#include <stdlib.h>

#include "typedef.h"
#include "basic_op.h"
#include "acelp.h"
#include "cnst.h"
#include "main.h"
#include "bits.h"
#include "dtx.h"
#include "count.h"

/*-------------------------------------------------------------------*
 * DECODER.C                                                         *
 * ~~~~~~~~~                                                         *
 * Main program of the AMR WB ACELP wideband decoder.				 *
 *                                                                   *
 *    Usage : decoder bitstream_file synth_file bit_rate             *
 *                                                                   *
 *    Format for bitstream_file:                                     *
 *        1 word (2-byte) for the type of frame type				 *
 *          (TX_FRAME_TYPE or RX_FRAME_TYPE) 						 *
 *        1 word (2-byte) for the frame type					     *
 *          (see dtx.h for possible values)						     *
 *        1 word (2-byte) for the mode indication				     *
 *          (see bits.h for possible values)					     *
 *		  N words (2-byte) containning N bits.                       *
 *          Bit 0 = 0x0000 and Bit 1 = 0x0001                        *
 *                                                                   *
 *    Format for synth_file:                                         *
 *      Synthesis is written to a binary file of 16 bits data.       *
 *                                                                   *
 *-------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
    FILE *f_serial;                        /* File of serial bits for transmission  */
    FILE *f_synth;                         /* File of speech data                   */

    Word16 synth[L_FRAME16k];              /* Buffer for speech @ 16kHz             */
    Word16 prms[NB_BITS_MAX];

    Word16 nb_bits, mode, frame_type, frame_length;
    Word16 reset_flag = 0;
    Word16 reset_flag_old = 1;
    Word16 mode_old = 0;
    Word16 i;
    long frame;

    void *st;


#ifdef ITU_interface

    Word16 quiet;

    quiet = 0;
    if(argc == 4)
    {
        if(strcmp(argv[1], "-q") == 0)
        {
            quiet = 1;
            argv++;
        }
        else
        {
            argc = 10;
        }
    }
     fprintf(stderr, " ========================================================================\n");
     fprintf(stderr, " 3GPP/ITU-T AMR Wideband Codec, November 30, 2001. Version %s.\n", CODEC_VERSION);
     fprintf(stderr, " (ITU-T interface).\n");
     fprintf(stderr, " ========================================================================\n");


    if ((argc < 3) || (argc > 4))
    {
        fprintf(stderr, "Usage : decoder  (-q) mode bitstream_file speech_file\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for speech_file:\n");
        fprintf(stderr, "  Speech is read form a binary file of 16 bits data.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for bitstream_file:\n");

        fprintf(stderr, "  One sync word (2-byte) (0x6b21 for good frames, 0x6b20 for bad frames).\n");
        fprintf(stderr, "  One word (2-byte) to indicate data length.\n");
        fprintf(stderr, "  N words (2-byte) containing N data bits.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "mode: 0 to 8 (9 bits rates)\n");
        fprintf(stderr, " ===================================================================\n");
        fprintf(stderr, " mode   :  (0)  (1)   (2)   (3)   (4)   (5)   (6)   (7)   (8)     \n");
        fprintf(stderr, " bitrate: 6.60 8.85 12.65 14.25 15.85 18.25 19.85 23.05 23.85 kbit/s\n");
        fprintf(stderr, " ===================================================================\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "-q is quiet operation\n");
        fprintf(stderr, "\n");
        exit(0);
    }

    /* Open file for synthesis and packed serial stream */

    if ((f_serial = fopen(argv[1], "rb")) == NULL)
    {
        fprintf(stderr, "Input file '%s' does not exist !!\n", argv[1]);
        exit(0);
    } else
    if(quiet == 0)
        fprintf(stderr, "Input bitstream file:   %s\n", argv[1]);

    if ((f_synth = fopen(argv[2], "wb")) == NULL)
    {
        fprintf(stderr, "Cannot open file '%s' !!\n", argv[2]);
        exit(0);
    } else
    if(quiet == 0)
        fprintf(stderr, "Synthesis speech file:   %s\n", argv[2]);

#else

      fprintf(stderr, "\n");
      fprintf(stderr, " ==========================================================================\n");
      fprintf(stderr, " 3GPP/ITU-T AMR Wideband Codec, November 30, 2001. Version %s.\n", CODEC_VERSION);
      fprintf(stderr, " ==========================================================================\n");
      fprintf(stderr, "\n");


    /*-----------------------------------------------------------------*
     *           Read passed arguments and open in/out files           *
     *-----------------------------------------------------------------*/

    if (argc != 3)
    {
        fprintf(stderr, "Usage : decoder  bitstream_file  synth_file\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for bitstream_file:\n");
        fprintf(stderr, "  One word (2-byte) to indicate type of frame type.\n");
        fprintf(stderr, "  One word (2-byte) to indicate frame type.\n");
        fprintf(stderr, "  One word (2-byte) to indicate mode.\n");
        fprintf(stderr, "  N words (2-byte) containing N bits.\n");
        fprintf(stderr, "\n");
        fprintf(stderr, "Format for synth_file:\n");
        fprintf(stderr, "  Synthesis is written to a binary file of 16 bits data.\n");
        fprintf(stderr, "\n");
        exit(0);
    }

#ifdef NO_BFH
	fprintf(stderr, "\n");
	fprintf(stderr, "\n---------------------------------------------------------------------- ");
    fprintf(stderr, "\nThis code was compiled with a flag NO_BFH. This flag turns off the     ");
    fprintf(stderr, "\nbad frame handling functionality of the codec. Purpose of this flag    ");
    fprintf(stderr, "\nis only to indicate the locations of code lines that are for the       ");
    fprintf(stderr, "\nthe bad frame substitution only. This version is NOT intended to be    ");
    fprintf(stderr, "\nused for decoding of AMR-WB speech samples.");
    fprintf(stderr, "\n(Please re-compile the code without NO_BFH flag to get working version)");
    fprintf(stderr, "\n---------------------------------------------------------------------- ");
    fprintf(stderr, "\n\n\n");
    exit(-1);
#endif



    /* Open file for synthesis and packed serial stream */

    if ((f_serial = fopen(argv[1], "rb")) == NULL)
    {
        fprintf(stderr, "Input file '%s' does not exist !!\n", argv[1]);
        exit(0);
    } else
        fprintf(stderr, "Input bitstream file:   %s\n", argv[1]);

    if ((f_synth = fopen(argv[2], "wb")) == NULL)
    {
        fprintf(stderr, "Cannot open file '%s' !!\n", argv[2]);
        exit(0);
    } else
        fprintf(stderr, "Synthesis speech file:   %s\n", argv[2]);

#endif
    /*-----------------------------------------------------------------*
     *           Initialization of decoder                             *
     *-----------------------------------------------------------------*/

    Init_decoder(&st);
    Init_WMOPS_counter();

    /*-----------------------------------------------------------------*
     *            Loop for each "L_FRAME" speech data                  *
     *-----------------------------------------------------------------*/

#ifdef ITU_interface
    if(quiet == 0)
#endif
    fprintf(stderr, "\n --- Running ---\n");

    frame = 0;
    while ((nb_bits = Read_serial(f_serial, prms, &frame_type, &mode)) != 0)
    {
        Reset_WMOPS_counter();

        frame++;
#ifdef ITU_interface
    if(quiet == 0)
#endif
        fprintf(stderr, " Frames processed: %ld\r", frame);

        if ((frame_type == RX_NO_DATA) | (frame_type == RX_SPEECH_LOST))
        {
           mode = mode_old;
           reset_flag = 0;
        }
        else
        {
           mode_old = mode;

        /* if homed: check if this frame is another homing frame */
        if (reset_flag_old == 1)
        {
            /* only check until end of first subframe */
            reset_flag = decoder_homing_frame_test_first(prms, mode);
        }
        }

        /* produce encoder homing frame if homed & input=decoder homing frame */
        if ((reset_flag != 0) && (reset_flag_old != 0))
        {
            for (i = 0; i < L_FRAME16k; i++)
            {
                synth[i] = EHF_MASK;
            }
        } else
        {
            decoder(mode, prms, synth, &frame_length, st, frame_type);
        }

        for (i = 0; i < L_FRAME16k; i++)   /* Delete the 2 LSBs (14-bit output) */
        {
            synth[i] = (Word16) (synth[i] & 0xfffC);      logic16(); move16();
        }

        fwrite(synth, sizeof(Word16), L_FRAME16k, f_synth);

        WMOPS_output((Word16) (mode == MRDTX));

        /* if not homed: check whether current frame is a homing frame */
        if (reset_flag_old == 0)
        {
            /* check whole frame */
            reset_flag = decoder_homing_frame_test(prms, mode);
        }
        /* reset decoder if current frame is a homing frame */
        if (reset_flag != 0)
        {
            Reset_decoder(st, 1);
        }
        reset_flag_old = reset_flag;

    }

    Close_decoder(st);

    exit(0);
}
